import sqlite3
from collections import namedtuple
from typing import Generator, NamedTuple, Optional

import notaorm
from notaorm.condition import Condition
from notaorm.sql import option, order

sqlite3.register_converter("BOOLEAN", lambda v: v.decode() == 'True')


class Query:
    def __init__(self, table_name, table_rows=(), init_condition=None):
        self.table_name = table_name
        self.init_condition = init_condition
        self.foreign_rows = [r for r in table_rows if type(r).__name__ is 'ForeignKey']
        self._conn = sqlite3.connect(notaorm.database, detect_types=sqlite3.PARSE_DECLTYPES)

        if notaorm.print_query:
            self._conn.set_trace_callback(print)

    def _get_table_object(self, descriptions: tuple):
        return namedtuple(self.table_name, [desc[0] for desc in descriptions])

    @staticmethod
    def _append_option(query: str, **kwargs):
        for key in kwargs.keys():
            if not hasattr(option, key.upper()):
                raise NotImplementedError('Option not implement')

        sorted_option = {k: v for k, v in sorted(kwargs.items(), key=lambda t: getattr(option, t[0].upper())[1])}
        for key, value in sorted_option.items():
            if not hasattr(option, key.upper()):
                raise NotImplementedError('Option not implement')
            append_option = getattr(option, key.upper())[0]

            if type(value) == list:
                append_option = append_option.format(*value)
            else:
                append_option = append_option.format(value)
            query += append_option

        return query

    def _set_relation(self, fetch, response):
        fetch = list(fetch)
        for row in self.foreign_rows:
            all_row = [desc[0] for desc in response.description]
            index_foreign_row = all_row.index(row.row_name)
            fetch[index_foreign_row] = Relation(fetch[index_foreign_row], row)

        return fetch

    def _fetch(self, query: str, columns, condition=None, **kwargs):
        if columns is not '*':
            columns = ','.join(repr(c) for c in columns) if type(columns) is list else repr(columns)

        condition_values = []
        if condition is not None:
            if self.init_condition is not None:
                condition &= self.init_condition
            condition_values = condition.values
            kwargs['condition'] = condition.left_side
        elif type(self.init_condition) is Condition:
            condition_values = self.init_condition.values
            kwargs['condition'] = self.init_condition.left_side

        full_query = self._append_option(query, **kwargs).replace('COLUMNS_NAME', columns)
        res = self.exec(full_query, *condition_values, commit=False)
        table_obj = self._get_table_object(res.description)

        return res, table_obj

    def _fetch_all(self, columns, **kwargs):
        res, table_obj = self._fetch(order.SELECT, columns, **kwargs)
        fetch = res.fetchall()

        for items in fetch:
            yield table_obj(*items)

    def fetch_one(self, columns, **kwargs):
        res, table_obj = self._fetch(order.SELECT, columns, **kwargs)
        fetch = res.fetchone()
        if fetch is None:
            return

        fetch = self._set_relation(fetch, res)
        return table_obj(*fetch)

    def exec(self, query: str, *args, commit=True):
        query = query.replace('TABLE_NAME', self.table_name)
        res = self._conn.execute(query, args)

        if commit:
            self._conn.commit()

        return res


class Change(Query):
    def update(self, condition, **columns) -> sqlite3.Cursor:
        columns_to_set = ','.join(f'{key} = ?' for key in columns.keys())
        values = list(columns.values()) + condition.values

        return self.exec(order.UPDATE.format(columns_to_set, condition.left_side), *values)

    def insert(self, **columns) -> sqlite3.Cursor:
        keys = ",".join(columns.keys())
        values = ','.join('?' * len(columns.values()))

        return self.exec(order.INSERT.format(keys, values), *columns.values())

    def delete(self, condition: Condition, commit=False) -> sqlite3.Cursor:
        return self.exec(order.DELETE.format(condition.left_side), *condition.values, commit=commit)


class Show(Query):
    def all(self, columns='*', **options) -> Generator:
        return self._fetch_all(columns, **options)

    def filter(self, condition: Condition, columns='*', **options) -> Generator:
        return self._fetch_all(columns, condition=condition, **options)

    def get(self, condition: Condition, columns='*', **options) -> Optional[NamedTuple]:
        return self.fetch_one(columns, condition=condition, **options)

    def first(self, columns='*') -> Optional[NamedTuple]:
        return self.fetch_one(columns, order_by_asc=f'{self.table_name}.ROWID', limit=1)

    def last(self, columns='*') -> Optional[NamedTuple]:
        return self.fetch_one(columns, order_by_desc=f'{self.table_name}.ROWID', limit=1)


class Relation(Show):
    def __init__(self, value, row):
        self._ref_table = row.references_table
        self.pk = value
        super().__init__(self._ref_table.table_name, self._ref_table.rows,
                         init_condition=(self._ref_table.pk == self.pk))

    def __repr__(self):
        return f'<Relation: {self._ref_table} {self.pk}>'



JOIN = ' JOIN {} on {} = {} ', 1
CONDITION = 'WHERE {}', 2
ORDER_BY = ' ORDER BY {}', 3
ORDER_BY_ASC = ' ORDER BY {} ASC ', 3
ORDER_BY_DESC = ' ORDER BY {} DESC ', 3
GROUP_BY = ' GROUP BY {} ', 4
LIMIT = ' LIMIT {} ', 5



from notaorm.condition import Condition
from notaorm.operator_enum import Comparator
from notaorm.sql import creation
from notaorm.table import _Row, MathFunction, Table


class _Generic(_Row):
    prefix: str

    def get_sql_creation(self):
        return super().get_sql_creation().format(prefix=self.prefix)


class _Number(_Generic):
    @property
    def sum(self):
        return MathFunction('SUM', self.__repr__(), f'sum_{self.row_name}')

    @property
    def avg(self):
        return MathFunction('AVG', self.__repr__(), f'avg_{self.row_name}')

    @property
    def min(self):
        return MathFunction('MIN', self.__repr__(), f'min_{self.row_name}')

    @property
    def max(self):
        return MathFunction('MAX', self.__repr__(), f'max_{self.row_name}')


class Int(_Number):
    prefix = creation.INTEGER

    def __init__(self, row_name: str, primary_key=False, **kwargs):
        super().__init__(row_name, **kwargs)
        self.primary_key = primary_key

    def get_sql_creation(self):
        if self.primary_key:
            self.prefix += f' {creation.PRIMARY_KEY} {creation.AUTO_INCREMENT}'
        return super().get_sql_creation().format(prefix=self.prefix)


class Float(_Number):
    prefix = creation.FLOAT


class Varchar(_Generic):
    def __init__(self, row_name, length=255, **kwargs):
        super().__init__(row_name, **kwargs)
        self.prefix = creation.VARCHAR.format(length)

    def like(self, search: str):
        return Condition(self.__repr__(), Comparator.str, '?', [search])

    def start_with(self, search: str):
        return self.like(f'{search}%')

    def end_with(self, search: str):
        return self.like(f'%{search}')

    def contain(self, search: str):
        return self.like(f'%{search}%')


class Text(_Generic):
    def __init__(self, row_name, length=5000, **kwargs):
        super().__init__(row_name, **kwargs)
        self.prefix = creation.TEXT.format(length)


class Date(_Generic):
    prefix = creation.DATE
    now = 'CURRENT_DATE '

    def __init__(self, row_name: str, default=None, **kwargs):
        kwargs['default'] = self.now if default == 'now' else f'`{default}`'
        super().__init__(row_name, **kwargs)


class Datetime(Date):
    prefix = creation.TIMESTAMP
    now = 'CURRENT_TIMESTAMP '


class Bool(_Generic):
    prefix = creation.BOOLEAN

    def __init__(self, row_name: str, default=None, **kwargs):
        if default is not None:
            kwargs['default'] = 1 if default else 0
        super().__init__(row_name, **kwargs)


class ForeignKey(_Generic):
    prefix = creation.FOREIGNKEY

    def __init__(self, row_name: str, reference: Table, **kwargs):
        self.references_table = reference
        self.prefix = self.prefix.format(row_name=row_name, *repr(reference.pk).replace("'", '').split('.'))
        super().__init__(row_name, **kwargs)



CREATE_TABLE = "create table TABLE_NAME({});"

NOT_NULL = 'NOT NULL '
DEFAULT = 'DEFAULT {} '
PRIMARY_KEY = 'constraint TABLE_NAME_pk primary key '
AUTO_INCREMENT = 'autoincrement '
UNIQUE = 'create unique index TABLE_NAME_{row_name}_uindex on "TABLE_NAME" ({row_name})'

INTEGER = 'INTEGER'
VARCHAR = 'VARCHAR({})'
TEXT = 'TEXT({})'
FLOAT = 'FLOAT'
DATE = 'DATE'
TIMESTAMP = 'TIMESTAMP'
BOOLEAN = 'BOOLEAN'
FOREIGNKEY = '{row_name} INTEGER, FOREIGN KEY({row_name}) REFERENCES {}({})'


SELECT = 'SELECT COLUMNS_NAME FROM `TABLE_NAME` '

UPDATE = 'UPDATE `TABLE_NAME` SET {} WHERE {} '
INSERT = 'INSERT INTO `TABLE_NAME` ({}) VALUES ({})'
DELETE = 'DELETE FROM `TABLE_NAME` WHERE {}'

SUM = 'SUM({}) as {}'
COUNT = 'COUNT({}) as {}'
MAX = 'MAX({}) as {}'
MIN = 'MIN({}) as {}'
AVG = 'AVG({}) as {}'

WHERE = '{} {} {}'


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="NotAOrm",
    python_requires='>=3.6',
    version="1.2.1",
    author="ludel",
    author_email="ludel47@gmail.com",
    description="A sample python library for managing a SQLite database",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ludel/NotAOrm",
    packages=['notaorm', 'notaorm/sql'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
